#include <usercore.h>
#include <stdio.h>
#include <math.h>

int pixwindd();
int bw2dd();
struct vwsurf vwsurf = DEFAULT_VWSURF(pixwindd);
static short raster_cell2[16] = {
  0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff,
0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff, 0Xffff};
static struct suncore_raster brush_raster = {16, 16, 1, raster_cell2};
static short raster_cell[16] = {
  0X0000, 0X0000, 0X0000, 0X0000, 0X0000, 0X0000, 0X0000, 0X0000,
0X0000, 0X0000, 0X0000, 0X0000, 0X0000, 0X0000, 0X0000, 0X0000};
static struct suncore_raster eraser_raster = {16, 16, 1, raster_cell};

#define RIGHT_BUTTON_MASK 1
#define CENTER_BUTTON_MASK 2
#define LEFT_BUTTON_MASK 4
#define BOX_SEGMENT 1
#define DRAWING_SEGMENT 2
#define CELL_SEGMENT 3
#define MIN_X 0.2		       /* could be as low as 0.0 */
#define MAX_X 0.7		       /* could be as high as 1.0 */
#define MIN_Y 0.1		       /* could be as low as 0.0 */
#define MAX_Y 0.6		       /* could be as high as 0.75 */
#define X_CENTER ((MAX_X+MIN_X)/2.0)
#define Y_CENTER ((MAX_Y+MIN_Y)/2.0)
#define NUM_CELLS	1	       /* number of cells on in each
				        * direction from raster -> cells */
#define SAMPLE_DENS 2
#define BACKGROUND 0
#define FOREGROUND 16
#define BLACK 16
#define GREY 8
static float box_x[] = {MAX_X, MAX_X, MIN_X, MIN_X};
static float box_y[] = {MIN_Y, MAX_Y, MAX_Y, MIN_Y};
static float cell_dx[4], cell_dy[4];
static float x_array[100], y_array[100];
static int i, j;
static float cell_x_size, cell_y_size, small_x, small_y;
static int button;
static float x, y, last_x, last_y;

/* global data */
int cell_max_i, cell_max_j;
int cell[32][32];
extern int resizing, centering;

typedef char bool;


my_init_things()
{

  /* first some cell things */
  cell_x_size = (MAX_X - MIN_X) / (float) cell_max_i;
  cell_y_size = (MAX_Y - MIN_Y) / (float) cell_max_j;
  small_x = 0.05 * cell_x_size;
  small_y = 0.05 * cell_y_size;
  cell_dx[0] = cell_x_size - 2 * small_x;
  cell_dx[1] = 0.;
  cell_dx[2] = -cell_x_size + 2 * small_x;
  cell_dx[3] = 0.;
  cell_dy[0] = 0.;
  cell_dy[1] = cell_y_size - 2 * small_y;
  cell_dy[2] = 0.;
  cell_dy[3] = -cell_y_size + 2 * small_y;

  /* now some graphics things */
  initialize_core(DYNAMICB, SYNCHRONOUS, TWOD);
  initialize_view_surface(&vwsurf, FALSE);
  select_view_surface(&vwsurf);
  initialize_device(LOCATOR, 1);
  set_echo_surface(LOCATOR, 1, &vwsurf);
  set_echo(LOCATOR, 1, 1);	       /* finger pointer */
  initialize_device(BUTTON, 1);
  initialize_device(BUTTON, 2);
  initialize_device(BUTTON, 3);
  set_echo_surface(BUTTON, 1, &vwsurf);
  set_echo_surface(BUTTON, 2, &vwsurf);
  set_echo_surface(BUTTON, 3, &vwsurf);

  create_retained_segment(DRAWING_SEGMENT);
}


my_init_cell_things()
{

  /* first some cell things */
  cell_x_size = (MAX_X - MIN_X) / (float) cell_max_i;
  cell_y_size = (MAX_Y - MIN_Y) / (float) cell_max_j;
  small_x = 0.05 * cell_x_size;
  small_y = 0.05 * cell_y_size;
  cell_dx[0] = cell_x_size - 2 * small_x;
  cell_dx[1] = 0.;
  cell_dx[2] = -cell_x_size + 2 * small_x;
  cell_dx[3] = 0.;
  cell_dy[0] = 0.;
  cell_dy[1] = cell_y_size - 2 * small_y;
  cell_dy[2] = 0.;
  cell_dy[3] = -cell_y_size + 2 * small_y;
}


get_cells()
{

  delete_all_retained_segments();
  create_retained_segment(DRAWING_SEGMENT);
  my_draw_grid();
  my_get_drawing();
  if (centering) {
    center_image();
  }
  if (resizing) {
    resize_image();
  }
}


draw_cells()
{

  delete_all_retained_segments();
  create_retained_segment(DRAWING_SEGMENT);
  my_draw_grid();
  my_draw_cell();
}


toggle_cell(i, j)
int i, j;
{

  if (cell[i][j]) {
    my_draw_single_cell(i, j, BACKGROUND);
    cell[i][j] = 0;
  }
  else {
    my_draw_single_cell(i, j, FOREGROUND);
    cell[i][j] = 1;
  }
}


my_terminate_things()
{

  delete_all_retained_segments();
  terminate_device(LOCATOR, 1);
  terminate_device(BUTTON, 1);
  terminate_device(BUTTON, 2);
  terminate_device(BUTTON, 3);
  deselect_view_surface(&vwsurf);
  terminate_core();
}


static my_draw_grid()
{

  /* Draw surrounding box */
  set_linewidth(0.5);
  move_abs_2(MIN_X, MIN_Y);
  polyline_abs_2(box_x, box_y, 4);
  move_abs_2(MIN_X, MAX_Y + 0.1);
  text("BUTTONS:  left = draw  middle = clear  right = exit");
  set_linewidth(0.0);
  for (x = MIN_X; x < MAX_X; x += cell_x_size) {
    move_abs_2(x, MIN_Y);
    line_abs_2(x, MAX_Y);
  }
  for (y = MIN_Y; y < MAX_Y; y += cell_y_size) {
    move_abs_2(MIN_X, y);
    line_abs_2(MAX_X, y);
  }
}


center_image()
{
  float x_left, x_right, y_top, y_bottom;
  float x_center, y_center;
  float x_shift, y_shift;


  find_box(&x_left, &x_right, &y_top, &y_bottom);
  draw_box(x_left, x_right, y_top, y_bottom);
  x_center = (x_right + x_left) / 2.0;
  y_center = (y_top + y_bottom) / 2.0;
  x_shift = X_CENTER - x_center;
  y_shift = Y_CENTER - y_center;
  /* handle discreteness problem */
  if (x_shift < 0.0)
    x_shift -= cell_x_size;
  if (y_shift < 0.0)
    y_shift -= cell_y_size;
  translate_cells((int) (x_shift / cell_x_size), (int) (y_shift / cell_y_size));
  draw_cells();
  find_box(&x_left, &x_right, &y_top, &y_bottom);
  draw_box(x_left, x_right, y_top, y_bottom);
}


resize_image()
{
  float x_left, x_right, y_top, y_bottom;
  float x_stretch, y_stretch, x_size, y_size;


  find_box(&x_left, &x_right, &y_top, &y_bottom);
  x_size = x_left - X_CENTER;
  x_stretch = (.9 * (MIN_X - X_CENTER)) / x_size;
  y_size = y_bottom - Y_CENTER;
  y_stretch = (.9 * (MIN_Y - Y_CENTER)) / y_size;
  stretch_cells(x_stretch, y_stretch);
  draw_cells();
  find_box(&x_left, &x_right, &y_top, &y_bottom);
  draw_box(x_left, x_right, y_top, y_bottom);
}


static draw_box(x_left, x_right, y_top, y_bottom)
float x_left, x_right, y_top, y_bottom;
{
  float box_x[4];
  float box_y[4];


  /* Draw surrounding box */
  box_x[0] = x_right;
  box_x[1] = x_right;
  box_x[2] = x_left;
  box_x[3] = x_left;
  box_y[0] = y_bottom;
  box_y[1] = y_top;
  box_y[2] = y_top;
  box_y[3] = y_bottom;
  set_linewidth(0.5);
  move_abs_2(x_left, y_bottom);
  polyline_abs_2(box_x, box_y, 4);
}


static find_box(x_left_pt, x_right_pt, y_top_pt, y_bottom_pt)
float *x_left_pt, *x_right_pt, *y_top_pt, *y_bottom_pt;
{
  int i, j, hit_flag;
  float x, y;


  /* find y_bottom */
  y = MIN_Y;
  for (j = 0; j < cell_max_j; j++) {
    hit_flag = FALSE;
    for (i = 0; i < cell_max_i; i++) {
      if (cell[i][j]) {
	hit_flag = TRUE;
	break;
      }
    }
    if (hit_flag) {
      *y_bottom_pt = y;
      break;
    }
    y += cell_y_size;
  }

  /* find y_top */
  y = MAX_Y;
  for (j = cell_max_j - 1; j >= 0; j--) {
    hit_flag = FALSE;
    for (i = 0; i < cell_max_i; i++) {
      if (cell[i][j]) {
	hit_flag = TRUE;
	break;
      }
    }
    if (hit_flag) {
      *y_top_pt = y;
      break;
    }
    y -= cell_y_size;
  }

  /* find x_left */
  x = MIN_X;
  for (i = 0; i < cell_max_i; i++) {
    hit_flag = FALSE;
    for (j = 0; j < cell_max_j; j++) {
      if (cell[i][j]) {
	hit_flag = TRUE;
	break;
      }
    }
    if (hit_flag) {
      *x_left_pt = x;
      break;
    }
    x += cell_x_size;
  }
  /* find x_right */
  x = MAX_X;
  for (i = cell_max_i - 1; i >= 0; i--) {
    hit_flag = FALSE;
    for (j = 0; j < cell_max_j; j++) {
      if (cell[i][j]) {
	hit_flag = TRUE;
	break;
      }
    }
    if (hit_flag) {
      *x_right_pt = x;
      break;
    }
    x -= cell_x_size;
  }
}


static translate_cells(delta_i, delta_j)
int delta_i, delta_j;
{
  int i, j;
  int cell_temp[32][32];


  for (i = 0; i < cell_max_i; ++i) {
    for (j = 0; j < cell_max_j; ++j) {
      cell_temp[(i + delta_i + cell_max_i) % cell_max_i][(j + delta_j + cell_max_j) % cell_max_j] = cell[i][j];
    }
  }
  for (i = 0; i < cell_max_i; ++i) {
    for (j = 0; j < cell_max_j; ++j) {
      cell[i][j] = cell_temp[i][j];
    }
  }
}


static stretch_cells(x_stretch, y_stretch)
float x_stretch, y_stretch;
{
  int i, j, i_new[SAMPLE_DENS], j_new[SAMPLE_DENS], l, k;
  int cell_temp[32][32];


  for (i = 0; i < cell_max_i; ++i) {
    for (j = 0; j < cell_max_j; ++j) {
      cell_temp[i][j] = FALSE;
    }
  }
  for (i = 0; i < cell_max_i; ++i) {
    for (j = 0; j < cell_max_j; ++j) {
      if (cell[i][j]) {
	map_coord(i, j, i_new, j_new, x_stretch, y_stretch);
	for (l = 0; l < SAMPLE_DENS; ++l) {
	  for (k = 0; k < SAMPLE_DENS; ++k) {
	    cell_temp[i_new[l]][j_new[k]] = TRUE;
	  }
	}
      }
    }
  }
  for (i = 0; i < cell_max_i; ++i) {
    for (j = 0; j < cell_max_j; ++j) {
      cell[i][j] = cell_temp[i][j];
    }
  }
}


map_coord(i, j, i_new, j_new, x_stretch, y_stretch)
int i, j, *i_new, *j_new;
float x_stretch, y_stretch;
{
  int i_center, j_center;
  int l, k;
  float x_dist, y_dist, step;


  i_center = cell_max_i / 2;
  j_center = cell_max_j / 2;
  step = 1.0 / (float) SAMPLE_DENS;
  for (l = 0; l < SAMPLE_DENS; ++l) {
    for (k = 0; k < SAMPLE_DENS; ++k) {
      if (i < i_center) {
	x_dist = cell_x_size * ((float) (i_center - i) + step * l);
	x_dist *= x_stretch;
	i_new[l] = i_center - (int) (x_dist / cell_x_size);
      }
      else {
	x_dist = cell_x_size * ((float) (i - i_center) + step * l);
	x_dist *= x_stretch;
	i_new[l] = i_center + (int) (x_dist / cell_x_size);
      }
      if (j < j_center) {
	y_dist = cell_y_size * ((float) (j_center - j) + step * k);
	y_dist *= y_stretch;
	j_new[k] = j_center - (int) (y_dist / cell_y_size);
      }
      else {
	y_dist = cell_y_size * ((float) (j - j_center) + step * k);
	y_dist *= y_stretch;
	j_new[k] = j_center + (int) (y_dist / cell_y_size);
      }
    }
  }
}


static my_get_drawing()
{
  int l, k;


  clear_cells();

  set_fill_index(FOREGROUND);

  /* let user "draw" things first */
  get_mouse_state(LOCATOR, 1, &last_x, &last_y, &button);
  while ((button & RIGHT_BUTTON_MASK) == 0) {
    get_mouse_state(LOCATOR, 1, &x, &y, &button);
    if (button & LEFT_BUTTON_MASK) {
      move_abs_2(x, y);
      put_raster(&brush_raster);
      i = (int) ((x - MIN_X) / cell_x_size);
      j = (int) ((y - MIN_Y) / cell_y_size);
      for (l = 0; l < NUM_CELLS; ++l) {
	for (k = 0; k < NUM_CELLS; ++k) {
	  if ((i + l < cell_max_i) && (j + k < cell_max_j)) {
	    cell[i + l][j + k] = TRUE;
	  }
	}
      }
    }
    if (button & CENTER_BUTTON_MASK) {
      delete_all_retained_segments();
      create_retained_segment(DRAWING_SEGMENT);
      my_draw_grid();
      clear_cells();
    }
    last_x = x;
    last_y = y;
  }
  while ((button & RIGHT_BUTTON_MASK) != 0) {
    get_mouse_state(LOCATOR, 1, &x, &y, &button);
  }

  delete_all_retained_segments();
  create_retained_segment(DRAWING_SEGMENT);
  my_draw_grid();
  my_draw_cell();

  /* now let user toggle cells */
  get_mouse_state(LOCATOR, 1, &last_x, &last_y, &button);
  while ((button & RIGHT_BUTTON_MASK) == 0) {
    get_mouse_state(LOCATOR, 1, &x, &y, &button);
    if (button & LEFT_BUTTON_MASK) {
      i = (int) ((x - MIN_X) / cell_x_size);
      j = (int) ((y - MIN_Y) / cell_y_size);
      if (cell[i][j] == 0) {
	my_draw_single_cell(i, j, FOREGROUND);
	cell[i][j] = TRUE;
      }
    }
    if (button & CENTER_BUTTON_MASK) {
      i = (int) ((x - MIN_X) / cell_x_size);
      j = (int) ((y - MIN_Y) / cell_y_size);
      if (cell[i][j] != 0) {
	my_draw_single_cell(i, j, BACKGROUND);
	cell[i][j] = FALSE;
      }
    }
    last_x = x;
    last_y = y;
  }
  while ((button & RIGHT_BUTTON_MASK) != 0) {
    get_mouse_state(LOCATOR, 1, &x, &y, &button);
  }
}


static clear_cells()
{
  /* clear cell array */
  for (i = 0; i < cell_max_i; i++)
    for (j = 0; j < cell_max_j; j++)
      cell[i][j] = FALSE;
}


static my_draw_cell()
{

  set_fill_index(FOREGROUND);
  x = MIN_X;
  for (i = 0; i < cell_max_i; i++) {
    y = MIN_Y;
    for (j = 0; j < cell_max_j; j++) {
      if (cell[i][j]) {
	move_abs_2(x + small_x, y + small_y);
	polygon_rel_2(cell_dx, cell_dy, 4);
      }
      y += cell_y_size;
    }
    x += cell_x_size;
  }
}


static my_draw_single_cell(i, j, fill_index)
int i, j, fill_index;
{

  set_fill_index(fill_index);
  x = MIN_X + cell_x_size * (float) i;
  y = MIN_Y + cell_y_size * (float) j;
  move_abs_2(x + small_x, y + small_y);
  polygon_rel_2(cell_dx, cell_dy, 4);
}


show_float_vector(v, n, provide_norm, norm)
float v[], norm;
int n;
bool provide_norm;
{
  float x, y, start_x, start_y, max_square_size, size;
  int i, j, sqrt_n, pick;
  static char string[100];


  delete_all_retained_segments();
  create_retained_segment(DRAWING_SEGMENT);
  sqrt_n = (int) (sqrt((float) n));
  if (sqrt_n * sqrt_n < n - sqrt_n)
    sqrt_n++;
  x = (float) (MAX_X - MIN_X);
  y = (float) (MAX_Y - MIN_Y);
  x = (x < y ? x : y);
  max_square_size = x / sqrt_n;
  start_x = MIN_X + ((MAX_X - MIN_X - max_square_size * sqrt_n) / 2);
  start_y = MIN_Y + ((MAX_Y - MIN_Y - max_square_size * sqrt_n) / 2);
  if (!provide_norm) {
    norm = 0;
    for (i = 0; i < n; i++)
      if (fabs(v[i]) > norm)
	norm = fabs(v[i]);
  }
  move_abs_2(0.0, MAX_Y);
  sprintf(string, "Scale: %4.2f is  ", norm);
  text(string);
  move_rel_2(0.0, -(0.05 + max_square_size));
  show_vec_box(max_square_size, BLACK);
  x = start_x;
  y = start_y;
  for (i = 0; i < n; i++) {
    move_abs_2(x, y);
    size = fabs(v[i] / norm) * max_square_size;
    move_rel_2((max_square_size - size) / 2, (max_square_size - size) / 2);
    show_vec_box(size, (v[i] >= 0 ? BLACK : GREY));
    y += max_square_size;
    if ((i % sqrt_n) == (sqrt_n - 1)) {
      y = start_y;
      x += max_square_size;
    }
  }
  move_abs_2(MIN_X, MAX_Y + 0.1);
  text("BUTTONS:  left or middle = pick square  right = exit");
  get_mouse_state(LOCATOR, 1, &x, &y, &button);
  while ((button & RIGHT_BUTTON_MASK) == 0) {
    get_mouse_state(LOCATOR, 1, &x, &y, &button);
    if ((button & LEFT_BUTTON_MASK) || (button & CENTER_BUTTON_MASK)) {
      i = (x - start_x) / max_square_size;
      if (i < 0 || i >= sqrt_n)
	continue;
      j = (y - start_y) / max_square_size;
      if (j < 0 || j >= sqrt_n)
	continue;
      pick = i * sqrt_n + j;
      if (pick < 0 || pick >= n)
	continue;
      else
	return (pick);
    }
  }
  return (-1);			       /* didn't pick any square */
}


show_vec_box(size, fill)
float size;
int fill;
{
  static float box_dx[4], box_dy[4];


  box_dx[0] = size;
  box_dy[0] = 0.0;
  box_dx[1] = 0.0;
  box_dy[1] = size;
  box_dx[2] = -size;
  box_dy[2] = 0.0;
  box_dx[3] = 0.0;
  box_dy[3] = -size;
  set_fill_index(fill);
  polygon_rel_2(box_dx, box_dy, 4);
}
